<?php

namespace QUI\Demodata;

use QUI;
use QUI\Bricks\Brick;
use QUI\Bricks\Manager;
use QUI\Events\Event;
use QUI\Exception;
use QUI\Projects\Project;
use QUI\Projects\Site;
use QUI\System\Log;

use function array_keys;
use function count;
use function file_exists;
use function file_get_contents;
use function is_array;
use function json_decode;
use function json_encode;
use function ltrim;
use function preg_match;
use function reset;
use function sprintf;
use function str_replace;
use function trim;

/**
 * Class DemoData
 *
 * @package QUI\Demodata
 */
class DemoData
{
    protected array $identifiers
        = [
            'sites' => [],
            'bricks' => [],
            'media' => []
        ];

    /** @var Project */
    protected Project $Project;
    public Event $Events;

    /**
     * DemoDataParser constructor.
     */
    public function __construct()
    {
        $this->Events = new Event();
    }

    /**
     * Applies the given demo data to the project
     *
     * @param Project $Project
     * @param $demoDataArray
     *
     * @throws Exception
     */
    public function apply(Project $Project, $demoDataArray): void
    {
        // import events
        if (isset($demoDataArray['events'])) {
            $events = $demoDataArray['events'];

            foreach ($events as $event => $callbacks) {
                foreach ($callbacks as $callback) {
                    $this->Events->addEvent($event, $callback);
                }
            }
        }

        // Create media area
        if (!empty($demoDataArray['projects'][0]['media'])) {
            $Media = new Media();
            $Media->createMediaArea($Project, $demoDataArray);

            $this->identifiers['media'] = $Media->getIdentifiers();
        }

        if (isset($demoDataArray['bricks'])) {
            $this->identifiers['bricks'] = $this->addBricks($Project, $demoDataArray['bricks']);
        }

        // Set the project settings and create its children sites
        if (isset($demoDataArray['projects'])) {
            $this->Project = $this->configureProject($Project, $demoDataArray['projects'][0]);
        }

        $this->replacePlaceholdersInBrickSettings();
        $this->replacePlaceholdersInSiteSettings();
        $this->replacePlaceholdersInProjectSettings();

        if (isset($demoDataArray['projects'][0]['sql'])) {
            foreach ($demoDataArray['projects'] as $projectEntry) {
                if (!isset($projectEntry['sql'])) {
                    continue;
                }

                foreach ($projectEntry['sql'] as $sqlFile) {
                    if (!file_exists($sqlFile)) {
                        continue;
                    }

                    $sql = file_get_contents($sqlFile);
                    $sql = $this->processPlaceholders($sql);

                    QUI::getDataBase()->getPDO()->exec($sql);
                }
            }
        }

        $this->Events->fireEvent('finish', [$this, $Project]);
    }

    /**
     * Sets the given project settings and creates the defined sites
     *
     * @param Project $Project
     * @param $projectData
     *
     * @return Project
     * @throws Exception
     */
    protected function configureProject(Project $Project, $projectData): Project
    {
        $sites = $projectData['sites'];
        $settings = $projectData['settings'];

        // Set the settings
        $ProjectsConfig = QUI::getConfig('etc/projects.ini.php');

        foreach ($settings as $key => $value) {
            $ProjectsConfig->set($Project->getName(), $key, $value);
        }

        $ProjectsConfig->save();
        // Workaround to force the project manager to reload the project data
        unset(QUI\Projects\Manager::$projects[$Project->getName()]);

        if (count($sites) > 0) {
            $identifier = array_keys($sites)[0];

            $this->identifiers['sites'][$identifier] = $this->configureStartpage($Project, reset($sites));
        }

        $this->Project = $Project;

        return $Project;
    }

    //////////////////////
    //   Sites
    //////////////////////
    #region Sites
    /**
     * Creates a site and its children
     *
     * @param Site $Parent
     * @param $siteData
     *
     *
     * @return int - Returns the created sites ID
     * @throws Exception
     */
    protected function createSite(Site $Parent, $siteData): int
    {
        $EditSite = $Parent->getEdit();
        $newSiteID = $EditSite->createChild(
            $siteData['attributes'],
            [],
            QUI::getUsers()->getSystemUser()
        );

        $NewSite = new Site\Edit($Parent->getProject(), $newSiteID);
        $NewSite->activate(QUI::getUsers()->getSystemUser());
        $NewSite->setAttributes($siteData['attributes']);
        $NewSite->save();

        // Create children
        if (!empty($siteData['children'])) {
            foreach ($siteData['children'] as $identifier => $childData) {
                $this->identifiers['sites'][$identifier] = $this->createSite($NewSite, $childData);
            }
        }

        // Add bricks to the site
        if (!empty($siteData['bricks'])) {
            $this->addBrickToSite($NewSite, $siteData['bricks']);
        }

        return $newSiteID;
    }

    /**
     * Creates and configures the start page
     *
     * @param Project $Project
     * @param $siteData
     *
     * @return bool|int
     * @throws Exception
     */
    protected function configureStartpage(Project $Project, $siteData): bool|int
    {
        $Site = $Project->firstChild();
        $Site = $Site->getEdit();

        // Create children
        if (!empty($siteData['children'])) {
            foreach ($siteData['children'] as $identifier => $childData) {
                $this->identifiers['sites'][$identifier] = $this->createSite($Project->firstChild(), $childData);
            }
        }

        if (isset($siteData['attributes'])) {
            foreach ($siteData['attributes'] as $attribute => $value) {
                $Site->setAttribute($attribute, $value);
            }
            $Site->unlockWithRights(); //workaround
            $Site->save(QUI::getUsers()->getSystemUser());
        }

        // Add bricks to the site
        if (!empty($siteData['bricks'])) {
            $this->addBrickToSite($Site, $siteData['bricks']);
        }

        return $Site->getId();
    }
    #endregion

    //////////////////////
    //   Bricks
    //////////////////////
    #region Bricks
    /**
     * Adds the defined bricks to QUIQQER
     *
     * @param Project $Project
     * @param $bricksData
     *
     * @return array
     * @throws Exception
     */
    protected function addBricks(Project $Project, $bricksData): array
    {
        $BrickManager = Manager::init();
        $createdBricks = [];

        foreach ($bricksData as $brickIdentifier => $brickParams) {
            $Brick = new Brick($brickParams['attributes']);
            $brickID = $BrickManager->createBrickForProject($Project, $Brick);

            // Set the attributes of the brick
            foreach ($brickParams['attributes'] as $key => $value) {
                if ($key === 'settings') {
                    continue;
                }

                $Brick->setAttribute($key, $value);
            }

            // Set the bricks settings
            if (!empty($brickParams['settings'])) {
                $Brick->setSettings($brickParams['settings']);
            }

            $brickParams['customfields'] = json_decode($brickParams['attributes']['customfields'], true);
            $BrickManager->saveBrick($brickID, $brickParams);
            $createdBricks[$brickIdentifier] = $brickID;
        }

        return $createdBricks;
    }

    /**
     * Adds the given bricks to the given site
     *
     * @param QUI\Interfaces\Projects\Site $Site
     * @param array $bricksData
     *
     * @throws Exception
     */
    protected function addBrickToSite(QUI\Interfaces\Projects\Site $Site, array $bricksData): void
    {
        $BrickManager = Manager::init();
        $siteBricks = [];

        foreach ($bricksData as $areaName => $areaBricks) {
            foreach ($areaBricks as $brickData) {
                if (!isset($this->identifiers['bricks'][$brickData['identifier']])) {
                    continue;
                }

                $quiBrickData = [
                    'brickId' => $this->identifiers['bricks'][$brickData['identifier']],
                    'customfields' => json_encode($brickData['settings']),
                    'uid' => ''
                ];

                $brickUid = $BrickManager->createUniqueSiteBrick($Site, $quiBrickData);
                $quiBrickData['uid'] = $brickUid;

                $siteBricks[$areaName][] = $quiBrickData;
            }
        }

        $Site->setAttribute('quiqqer.bricks.areas', json_encode($siteBricks));
        $Site->save(QUI::getUsers()->getSystemUser());
    }
    #endregion

    //////////////////////
    //   Placeholders
    //////////////////////
    #region Placeholders
    /**
     * Replaces all placeholders within the bricks settings
     */
    protected function replacePlaceholdersInBrickSettings(): void
    {
        /*
         * Global brick settings
         */
        $BrickManager = new Manager();
        $Bricks = $BrickManager->getBricksFromProject($this->Project);

        /** @var Brick $Brick */
        foreach ($Bricks as $Brick) {
            $updatedSettings = [];
            $updatedAttributes = [];

            foreach ($Brick->getSettings() as $settingName => $settingValue) {
                $updatedSettings[$settingName] = $this->processPlaceholders($settingValue);
            }

            foreach ($Brick->getAttributes() as $settingName => $settingValue) {
                $updatedAttributes[$settingName] = $this->processPlaceholders($settingValue);
            }

            $Brick->setSettings($updatedSettings);
            $Brick->setAttributes($updatedAttributes);

            $BrickManager->saveBrick(
                $Brick->getAttribute('id'),
                $Brick->getAttributes()
            );
        }

        /*
         * Site specific settings
         */
        /** @var Site\Edit $Site */
        foreach ($this->Project->getSites() as $Site) {
            $siteBricks = json_decode($Site->getAttribute('quiqqer.bricks.areas'), true);

            if (empty($siteBricks)) {
                continue;
            }

            $updatedSiteBricks = [];

            foreach ($siteBricks as $brickArea => $bricks) {
                foreach ($bricks as $brickData) {
                    $brickSettings = json_decode($brickData['customfields'], true);
                    $updatedSettings = [];

                    foreach ($brickSettings as $settingName => $settingValue) {
                        $updatedSettings[$settingName] = $this->processPlaceholders($settingValue);
                    }

                    $brickData['customfields'] = json_encode($updatedSettings);
                    $updatedSiteBricks[$brickArea][] = $brickData;
                }
            }

            /*
             * Brick UID Table
             */
            $brickUIDRows = QUI::getDataBase()->fetch([
                'from' => Manager::getUIDTable(),
                'where' => [
                    'project' => $this->Project->getName(),
                    'lang' => $this->Project->getLang()
                ]
            ]);

            foreach ($brickUIDRows as $brickUIDRow) {
                $customSettings = json_decode($brickUIDRow['customfields'], true);

                foreach ($customSettings as $settingName => $settingValue) {
                    $customSettings[$settingName] = $this->processPlaceholders($settingValue);
                }

                QUI::getDataBase()->update(
                    Manager::getUIDTable(),
                    ['customfields' => json_encode($customSettings)],
                    ['uid' => $brickUIDRow['uid']]
                );
            }

            $Site->setAttribute('quiqqer.bricks.areas', $updatedSiteBricks);
            $Site->save();
        }
    }

    /**
     * Replaces all placeholders within all sites settings
     */
    protected function replacePlaceholdersInSiteSettings(): void
    {
        /** @var Site\Edit $Site */
        foreach ($this->Project->getSites() as $Site) {
            $updatedSettings = [];

            foreach ($Site->getAttributes() as $attributeName => $attributeValue) {
                $updatedSettings[$attributeName] = $this->processPlaceholders($attributeValue);
            }

            $Site->setAttributes($updatedSettings);
            $Site->save();
        }
    }

    /**
     * Replaces all Placeholders within the projects settings
     *
     * @throws Exception
     */
    protected function replacePlaceholdersInProjectSettings(): void
    {
        $ProjectsConfig = QUI::getConfig('etc/projects.ini.php');
        $settings = $ProjectsConfig->getSection($this->Project->getName());

        foreach ($settings as $key => $value) {
            $value = $this->processPlaceholders($value);
            $ProjectsConfig->set($this->Project->getName(), $key, $value);
        }

        $ProjectsConfig->save();
    }

    /**
     * Replaces all placeholders within the given string.
     * If an array is given, the method will convert the array into a json string and return the updated json string
     *
     * @param mixed $string
     * @return array|string
     */
    protected function processPlaceholders(mixed $string): array|string
    {
        if (empty($string)) {
            return $string;
        }

        if (is_bool($string)) {
            return '';
        }

        $placeholderPattern = [
            '/.*\$\{(project)\}.*/',
            '/.*\$\{(site\.)([a-zA-Z0-9\.\-_]+)\}.*/',
            '/.*\$\{(site):([a-zA-Z0-9\.\-_\/]+)\}.*/',
            '/.*\$\{(siteurl):([a-zA-Z0-9\.\-_\/]+)\}.*/',
            '/.*\$\{(media):([a-zA-Z0-9\.\-_\/\\\\]+)\}.*/',
        ];

        if (is_array($string)) {
            $string = json_encode($string);
        }

        foreach ($placeholderPattern as $pattern) {
            $string = $this->parsePlaceholders($pattern, $string);
        }

        return $string;
    }

    /**
     * @param $pattern
     * @param $string
     *
     * @return mixed
     */
    protected function parsePlaceholders($pattern, $string): mixed
    {
        preg_match($pattern, $string, $matches);

        if (isset($matches[1]) && $matches[1] === 'project') {
            $string = str_replace(
                '${project}',
                $this->Project->getName(),
                $string
            );

            return $this->parsePlaceholders($pattern, $string);
        }

        if (count($matches) !== 3) {
            return $string;
        }

        $type = $matches[1];
        $identifier = $matches[2];

        switch ($type) {
            case 'project':
                $string = str_replace(
                    '${project}',
                    $this->Project->getName(),
                    $string
                );

                break;

            case 'site.':
                if (!isset($this->identifiers['sites'][$identifier])) {
                    Log::addDebug('Site Identifier "' . $identifier . '" was not found');
                    break;
                }

                $string = str_replace(
                    '${site.' . $identifier . '}',
                    $this->identifiers['sites'][$identifier],
                    $string
                );
                break;

            case 'site':
                if (!isset($this->identifiers['sites'][$identifier])) {
                    Log::addDebug('Site Identifier "' . $identifier . '" was not found');
                    break;
                }

                $string = str_replace(
                    '${site:' . $identifier . '}',
                    $this->identifiers['sites'][$identifier],
                    $string
                );
                break;

            case 'siteurl':
                if (!isset($this->identifiers['sites'][$identifier])) {
                    Log::addDebug('Site Identifier "' . $identifier . '" was not found');
                    break;
                }

                $siteID = $this->identifiers['sites'][$identifier];
                $string = str_replace(
                    '${siteurl:' . $identifier . '}',
                    sprintf(
                        'index.php?id=%d&project=%s&lang=%s',
                        $siteID,
                        $this->Project->getName(),
                        $this->Project->getLang()
                    ),
                    $string
                );
                break;

            case 'media':
                $identifier = trim($identifier);
                $identifier = ltrim($identifier, '/ ');

                if (!isset($this->identifiers['media'][$identifier])) {
                    Log::addDebug('Media Identifier "' . $identifier . '" was not found');
                    break;
                }

                $replace = 'image.php?' .
                    'id=' . $this->identifiers['media'][$identifier] .
                    '&project=' . $this->Project->getName();

                $string = str_replace(
                    '${media:' . $identifier . '}',
                    $replace,
                    $string
                );
                break;
        }

        return $this->parsePlaceholders($pattern, $string);
    }
    #endregion
}
